//
// The provider for a tree of ECMA documents
//
// Authors:
//   Miguel de Icaza (miguel@ximian.com)
//   Joshua Tauberer (tauberer@for.net)
//
// (C) 2002, 2003 Ximian, Inc.
// (C) 2003 Joshua Tauberer.
//
// TODO:
//   Should cluster together constructors
//   Should cluster together methods with the same name (overloads)
//
// Easy:
//   Should render attributes on the signature.
//   Include examples as well.
//
using System;
using System.Reflection;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Text;
using System.Collections;
using ICSharpCode.SharpZipLib.Zip;

using BF = System.Reflection.BindingFlags;

//
// Helper routines to extract information from an Ecma XML document
//
public class EcmaDoc {
	public static string GetFullClassName (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type").Attributes ["FullName"].InnerText;
	}
	
	public static string GetClassName (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type").Attributes ["Name"].InnerText;
	}

	public static string GetClassAssembly (XmlDocument doc)
	{
		return doc.SelectSingleNode ("/Type/AssemblyInfo/AssemblyName").InnerText;
	}

	public static string GetClassNamespace (XmlDocument doc)
	{
		string s = doc.SelectSingleNode ("/Type").Attributes ["FullName"].InnerText;

		return s.Substring (0, s.LastIndexOf ("."));
	}
	
	public static string GetTypeKind (XmlDocument doc)
	{
		XmlNode node = doc.SelectSingleNode ("/Type/Base/BaseTypeName");

		if (node == null){
			if (GetFullClassName (doc) == "System.Object")
				return "Class";
			return "Interface";
		}

		switch (node.InnerText){

		case "System.Delegate":
			return "Delegate";
		case "System.ValueType":
			return "Structure";
		case "System.Enum":
			return "Enumeration";
		default:
			return "Class";
		}
	}

	//
	// Utility function: converts a fully .NET qualified type name into a C#-looking one
	//
	public static string ConvertCTSName (string type)
	{
		if (!type.StartsWith ("System."))
			return type;

		if (type.EndsWith ("*")) return ConvertCTSName(type.Substring(0, type.Length - 1)) + "*";
		if (type.EndsWith ("&")) return ConvertCTSName(type.Substring(0, type.Length - 1)) + "&";
		if (type.EndsWith ("[]")) return ConvertCTSName(type.Substring(0, type.Length - 2)) + "[]";

		switch (type) {
		case "System.Byte": return "byte";
		case "System.SByte": return "sbyte";
		case "System.Int16": return "short";
		case "System.Int32": return "int";
		case "System.Int64": return "long";
			
		case "System.UInt16": return "ushort";
		case "System.UInt32": return "uint";
		case "System.UInt64": return "ulong";
			
		case "System.Single":  return "float";
		case "System.Double":  return "double";
		case "System.Decimal": return "decimal";
		case "System.Boolean": return "bool";
		case "System.Char":    return "char";
		case "System.String":  return "string";
			
		case "System.Object":  return "object";
		case "System.Void":  return "void";
		}

		if (type.LastIndexOf(".") == 6) return type.Substring(7);
		
		return type;
	}
}

//
// The Ecma documentation provider:
//
// It populates a tree with contents during the help assembly
//
public class EcmaProvider : Provider {
	string basedir;

	public EcmaProvider (string base_directory)
	{
		basedir = base_directory;

		if (!Directory.Exists (basedir))
			throw new FileNotFoundException (String.Format ("The directory `{0}' does not exist", basedir));
	}
	
	public override void PopulateTree (Tree tree)
	{
		string [] namespaces = Directory.GetDirectories (basedir);

		foreach (string ns in namespaces){
			string [] files = Directory.GetFiles (ns);
			Node ns_node = null;
			string tn = null;

			foreach (string file in files){
				if (!file.EndsWith (".xml"))
					continue;

				if (ns_node == null){
					tn = Path.GetFileName (ns);
					Console.Error.WriteLine ("Processing namespace {0}", tn);
					ns_node = tree.LookupNode (tn, "N:" + tn);
					string ns_summary_file = basedir + "/" + tn + ".xml";
					if (File.Exists (ns_summary_file))
						ns_node.tree.HelpSource.PackFile (ns_summary_file,tn+".xml");
				}
				Console.Error.WriteLine ("    Processing input file {0}", Path.GetFileName (file));

				PopulateClass (tn, ns_node, file);
			}
		}
	}

	struct TypeInfo : IComparable {
		public string type_assembly;
		public string type_name;
		public string type_full;
		public string type_kind;
		public XmlNode type_doc;

		public TypeInfo (string k, string a, string f, string s, XmlNode n)
		{
			type_assembly = a;
			type_name = s;
			type_doc = n;
			type_kind = k;
			type_full = f;
		}

		public int CompareTo (object b)
		{
			TypeInfo na = this;
			TypeInfo nb = (TypeInfo) b;

			return String.Compare (na.type_full, nb.type_full);
		}
	}
	
	//
	// Packs a file with the summary data
	//
	public override void CloseTree (HelpSource hs, Tree tree)
	{
		foreach (DictionaryEntry de in class_summaries){
			XmlDocument doc = new XmlDocument ();
			string ns = (string) de.Key;
			
			ArrayList list = (ArrayList) de.Value;
			list.Sort();

			XmlElement elements = doc.CreateElement ("elements");
			doc.AppendChild (elements);
			
			string file = "xml.summary." + ns;
			Console.Error.WriteLine ("Have {0} elements in the {1}", list.Count, ns);
			foreach (TypeInfo p in list){
				XmlElement e = null;
				
				switch (p.type_kind){
				case "Class":
					e = doc.CreateElement ("class"); 
					break;
					
				case "Enumeration":
					e = doc.CreateElement ("enum");
					break;
					
				case "Structure":
					e = doc.CreateElement ("struct");
					break;
					
				case "Delegate":
					e = doc.CreateElement ("delegate");
					break;
					
				case "Interface":
					e = doc.CreateElement ("interface");
					break;
				}
				
				e.SetAttribute ("name", p.type_name);
				e.SetAttribute ("fullname", p.type_full);
				e.SetAttribute ("assembly", p.type_assembly);
				XmlNode copy = doc.ImportNode (p.type_doc, true);
				e.AppendChild (copy);
				elements.AppendChild (e);
			}
			hs.PackXml ("xml.summary." + ns, doc);
		}
	}
	       
	static Hashtable class_summaries = new Hashtable ();
					     
	XmlDocument doc;
	
	void PopulateClass (string ns, Node ns_node, string file)
	{
		doc = new XmlDocument ();
		doc.Load (file);
		
		string name = EcmaDoc.GetClassName (doc);
		string assembly = EcmaDoc.GetClassAssembly (doc);
		string kind = EcmaDoc.GetTypeKind (doc);
		string full = EcmaDoc.GetFullClassName (doc);

		Node class_node;
		string file_code = ns_node.tree.HelpSource.PackFile (file);

		XmlNode class_summary = doc.SelectSingleNode ("/Type/Docs/summary");
		ArrayList l = (ArrayList) class_summaries [ns];
		if (l == null)
			l = class_summaries [ns] = new ArrayList ();
		l.Add (new TypeInfo (kind, assembly, full, name, class_summary));
	       
		class_node = ns_node.LookupNode (String.Format ("{0} {1}", name, kind), "ecma:" + file_code + "#" + name + "/");

		if (kind == "Enumeration")
			return;

		if (kind == "Delegate")
			return;
		
		//
		// Always add the Members node
		//
		class_node.CreateNode ("Members", "*");

		PopulateMember (name, class_node, "Constructor", "Constructors");
		PopulateMember (name, class_node, "Method", "Methods");
		PopulateMember (name, class_node, "Property", "Properties");
		PopulateMember (name, class_node, "Field", "Fields");
		PopulateMember (name, class_node, "Event", "Events");
	}

	class NodeIndex {
		public XmlNode node;
		public int     index;

		public NodeIndex (XmlNode node, int index)
		{
			this.node = node;
			this.index = index;
		}
	}

	struct NodeCompare : IComparer {
		public int Compare (object a, object b)
		{
			NodeIndex na = (NodeIndex) a;
			NodeIndex nb = (NodeIndex) b;

			return String.Compare (na.node.Attributes ["MemberName"].InnerText,
					       nb.node.Attributes ["MemberName"].InnerText);
		}
	}

	static NodeCompare NodeComparer = new NodeCompare ();

	string GetMemberName (XmlNode node)
	{
		return node.Attributes ["MemberName"].InnerText;
	}
	
	//
	// Performs an XPath query on the document to extract the nodes for the various members
	// we also use some extra text to pluralize the caption
	//
	void PopulateMember (string typename, Node node, string text, string plural_text)
	{
		XmlNodeList list = doc.SelectNodes (String.Format ("/Type/Members/Member[MemberType=\"{0}\"]", text));
		int count = list.Count;
		
		if (count == 0)
			return;

		Node nodes_node;
		string key = text.Substring (0, 1);
		if (count == 1)
			nodes_node = node.CreateNode (text, key);
		else
			nodes_node = node.CreateNode (plural_text, key);

		NodeIndex [] node_array = new NodeIndex [count];
		int i = 0;
		foreach (XmlNode n in list)
			node_array [i] = new NodeIndex (n, i++);
			
		Array.Sort (node_array, NodeComparer);

		foreach (NodeIndex ni in node_array){
			string signature;

			switch (text){
			case "Event":
			case "Property":
			case "Field":
				signature = GetMemberName (ni.node);
				break;

			case "Constructor":
				signature = EcmaHelpSource.MakeSignature(ni.node, typename);
				break;

			case "Method":
				signature = EcmaHelpSource.MakeSignature(ni.node, null);
				break;
				
			default:
				XmlNode signode = ni.node.SelectSingleNode ("MemberSignature[@Language='C#']");
			
				if (signode == null)
					signature = GetMemberName (ni.node);
				else
					signature = signode.Attributes ["Value"].InnerText;
				break;
			}
			
			//signature = CleanSignature (ni.node, signature);

			nodes_node.CreateNode (signature, ni.index.ToString ());
		}
	}

}

//
// The Ecma HelpSource provider
//
public class EcmaHelpSource : HelpSource {

	public EcmaHelpSource (string base_file, bool create) : base (base_file, create)
	{
	}

	public override string GetText (string url, out Node match_node)
	{
		match_node = null;
		if (url.StartsWith ("ecma:"))
			return GetTextFromUrl (url);

		return null;
	}


	string RenderMemberLookup (string typename, string member, ref Node type_node)
	{
		if (type_node.Nodes == null) return null;

		string lasturl = "";
		XmlNode doc = null;

		string membername = member;
		string[] argtypes = null;
		if (member.IndexOf("(") > 0) {
			membername = membername.Substring(0, member.IndexOf("("));
			member = member.Replace("@", "&");
			
			// reform the member signature with CTS names

			string x = member.Substring(member.IndexOf("(")+1);
			argtypes = x.Substring(0, x.Length-1).Split(',');

			if (membername == ".ctor")
				member = typename;
			else
				member = membername;

			member += "(";
			for (int i = 0; i < argtypes.Length; i++) {
				argtypes[i] = EcmaDoc.ConvertCTSName(argtypes[i]);
				if (i > 0) member += ",";
				member += argtypes[i];
			}
			member += ")";
		}

		// Check if a node caption matches exactly

		foreach (Node x in type_node.Nodes){
			if (x.Nodes == null) continue;
			foreach (Node m in x.Nodes) {
				if (m.Caption == member || m.Caption.StartsWith(member + "(")) {
					type_node = m;
					return GetTextFromUrl (m.URL);
				}
			}
		}

		// Check the signatures within each document (sllloooow)
			
		foreach (Node x in type_node.Nodes){
			if (x.Nodes == null) continue;
			foreach (Node m in x.Nodes){
				if (m.Caption.StartsWith(membername) || membername == ".ctor" || (membername == "Finalize" && m.Caption.StartsWith("~")) || (membername.StartsWith("op_") && m.Caption.IndexOf("operator") >= 0) || m.Caption.IndexOf("this") >= 0) {

					string url = m.URL.Substring (5);

					string rest;
					string file = GetFile (url, out rest);

					if (url != lasturl) {
						doc = GetDocument (this, file).DocumentElement;
						lasturl = url;
					}

					string [] nodes = rest.Split (new char [] {'/'});
					int idx = int.Parse(nodes [1]);
					string nodetype = "";

					switch (nodes [0]){
					case "C": nodetype = "Constructor"; break;
					case "M": nodetype = "Method"; break;
					case "P": nodetype = "Property"; break;
					case "F": nodetype = "Field"; break;
					case "E": nodetype = "Event"; break;
					}

					XmlNode n = doc.SelectNodes("Members/Member[MemberType='" + nodetype + "']")[idx];

					if (n == null) continue;

					string sig = MakeSignature(n, nodes[0] == "C" ? typename : null);

					//Console.WriteLine(sig + " -- " + member);

					if (sig == member) {
						type_node = m;
						return GetTextFromUrl (m.URL);
					}
				}
			}
		}

		return null;
	}

	public static string MakeSignature(XmlNode n, string cstyleclass) {
		string sig;

		if (cstyleclass == null)
			sig = n.Attributes["MemberName"].InnerText;
		else // constructor style
			sig = cstyleclass;
	
		if (n.SelectSingleNode("MemberType").InnerText == "Method" || n.SelectSingleNode("MemberType").InnerText == "Constructor") {
			XmlNodeList paramnodes = n.SelectNodes("Parameters/Parameter");
			sig += "(";
			bool first = true;
			foreach (XmlNode p in paramnodes) {
				if (!first) sig += ",";
				string type = p.Attributes["Type"].InnerText;
				type = EcmaDoc.ConvertCTSName(type);
				sig += type;
				first = false;
			}
			sig += ")";
		}

		return sig;
	}
	
	//
	// This routine has to perform a lookup on a type.
	//
	// Example: T:System.Text.StringBuilder
	//
	// The prefix is the kind of opereation being requested (T:, E:, M: etc)
	// ns is the namespace being looked up
	// type is the type being requested
	//
	// This has to walk our toplevel (which is always a namespace)
	// And then the type space, and then walk further down depending on the request
	//
	public override string RenderTypeLookup (string prefix, string ns, string type, string member, out Node match_node)
	{
			
		// If a nested type, compare only inner type name to node list.
		// This should be removed when the node list doesn't lose the continaing type name.
		if (type.IndexOf("+") > 0) type = type.Substring(type.IndexOf("+")+1);

		string nsp = prefix + ns;

		foreach (Node ns_node in Tree.Nodes){
			string ns_node_namespace = ns_node.Element.Substring (2);

			if (ns_node_namespace != ns)
				continue;
			
			foreach (Node type_node in ns_node.Nodes){
				string element = type_node.Element;
				int pidx = element.IndexOf ("#");
				int sidx = element.IndexOf ("/");
				
				string cname = element.Substring (pidx + 1, sidx-pidx-1);
				
				//Console.WriteLine ("t:{0} cn:{1} p:{2}", type, cname, prefix);

				if (type == cname && prefix == "T:"){
					match_node = type_node;
					return GetTextFromUrl (type_node.URL);
				} else if (type.StartsWith (cname)){
					int p = cname.Length;

					match_node = type_node;
					if (type == cname){
						string ret = RenderMemberLookup (type, member, ref match_node);
						if (ret == null) return GetTextFromUrl (type_node.URL);
						return ret;

					} else if (type [p] == '/'){
						//
						// This handles summaries
						//
						match_node = null;
						Console.Error.WriteLine ("FIXME: Should set the right node here");
						return GetTextFromUrl (type_node.URL + type.Substring (p + 1));
					}
				}
			}
			match_node = null;
			return String.Format ("<b>FOUND!</b><p><p><b>Namespace: <b>{0}<br><b>Type:</b> {1}", ns, type);
		}

		match_node = null;
		return null;
	}


	public override string RenderNamespaceLookup (string nsurl, out Node match_node)
	{
		foreach (Node ns_node in Tree.Nodes){
			if (ns_node.Element != nsurl)
				continue;

			match_node = ns_node;
			StringBuilder sb = new StringBuilder ();
			string ns_name = nsurl.Substring (2);
			Stream intro = GetHelpStream (ns_name + ".xml");
			
			Stream summary = GetHelpStream ("xml.summary." + ns_name);
			if (summary == null)
				return null;
			
			XmlDocument summary_doc = new XmlDocument ();
			summary_doc = new XmlDocument ();
			summary_doc.Load (summary);
	
			if (intro != null){
				XmlDocument intro_doc = new XmlDocument ();
				intro_doc.Load (intro);

				summary_doc.DocumentElement.AppendChild( summary_doc.ImportNode( intro_doc.SelectSingleNode ("Namespace/Docs/summary"), true));

				summary_doc.DocumentElement.AppendChild( summary_doc.ImportNode( intro_doc.SelectSingleNode ("Namespace/Docs/remarks"), true));
			}

			XsltArgumentList args = new XsltArgumentList();
			args.AddExtensionObject("monodoc://extensions", ExtObject);
			args.AddParam("show", "", "namespace");
			args.AddParam("namespace", "", ns_name);
			return Htmlize(summary_doc, args);

		}
		match_node = null;
		return null;
	}

							   
	string GetTextFromUrl (string url) {
		XmlDocument doc;
		string file, rest;

		//
		// Remove ecma:
		//
		url = url.Substring (5);
		file = GetFile (url, out rest);

		// Console.WriteLine ("Got [{0}] and [{1}]", file, rest);
		doc = GetDocument (this, file);
		XmlNode docnode = doc; //.DocumentElement;

		XsltArgumentList args = new XsltArgumentList();

		args.AddExtensionObject("monodoc://extensions", ExtObject);
		
		if (rest == "") {
			args.AddParam("show", "", "typeoverview");
			return Htmlize(docnode, args);
		}
		
		string [] nodes = rest.Split (new char [] {'/'});
		if (nodes.Length == 1) {
			args.AddParam("show", "", "members");
			args.AddParam("index", "", "all");
		} else {
			args.AddParam("show", "", "member");
			args.AddParam("index", "", nodes [1]);
		}

		switch (nodes [0]){
		case "C":
			args.AddParam("membertype", "", "Constructor");
			break;
		case "M":
			args.AddParam("membertype", "", "Method");
			break;
		case "P":
			args.AddParam("membertype", "", "Property");
			break;
		case "F":
			args.AddParam("membertype", "", "Field");
			break;
		case "E":
			args.AddParam("membertype", "", "Event");
			break;
		case "*":
			args.AddParam("membertype", "", "All");
			break;
		default:
			return "Unknown url: " + url;
		}

		return Htmlize(docnode, args);
	}

	
	

	static XslTransform ecma_transform;

	static string Htmlize (XmlNode ecma_xml)
	{
		return Htmlize(ecma_xml, null);
	}

	static string Htmlize (XmlNode ecma_xml, XsltArgumentList args)
	{
		if (ecma_transform == null){
			ecma_transform = new XslTransform ();
			Assembly assembly = System.Reflection.Assembly.GetCallingAssembly ();
			Stream stream = assembly.GetManifestResourceStream ("mono-ecma.xsl");

			XmlReader xml_reader = new XmlTextReader (stream);
			ecma_transform.Load (xml_reader);
		}
		
		if (ecma_xml == null) return "";

		XmlDocument new_root;

		if (ecma_xml is XmlDocument) {
			new_root = (XmlDocument)ecma_xml;
		} else {
			new_root = new XmlDocument ();
			new_root.AppendChild (new_root.ImportNode (ecma_xml, true));
			//new_root.Save ("/tmp/blah.xml");
		}

		StringWriter output = new StringWriter ();
		ecma_transform.Transform (new_root, args, output);
		
		return output.ToString ();
	}

	// This ExtensionObject stuff is used to check at run time whether
	// types and members have been implemented and whether there are any
	// MonoTODO attributes left on them. 

	private readonly ExtensionObject ExtObject = new ExtensionObject();
	private class ExtensionObject {
		public string Test() { return "TestReturnText!"; }

		// Used by stylesheet to nicely reformat the <see cref=> tags. 
		public string MakeNiceSignature(string sig) {
			if (sig.Length < 3) return sig;
			if (sig[1] != ':') return sig;

			char s = sig[0];
			sig = sig.Substring(2);

			switch (s) {
				case 'N': return sig;
				case 'T': return ShortTypeName(sig);

				case 'C': case 'M': case 'P': case 'F': case 'E':
					int paren = sig.IndexOf("(");
					if (paren > 0) {
						string name = sig.Substring(0, paren);
						string[] args = sig.Substring(paren, sig.Length-paren-1).Split(',');
						for (int i = 0; i < args.Length; i++)
							args[i] = ShortTypeName(args[i]);
						return ShortTypeName(name) + "(" + String.Join(", ", args) + ")";
					} else {
						return ShortTypeName(sig);
					}

				default:
					return sig;
			}
		}

		private static string ShortTypeName(string name) {
			if (name.LastIndexOf(".") > 0) name = name.Substring(name.LastIndexOf(".")+1);
			return name;
		}


		public string MonoImpInfo(string assemblyname, string typename, string membername, string arglist, bool strlong) {
			ArrayList a = new ArrayList();
			if (arglist != "") a.Add(arglist);
			return MonoImpInfo(assemblyname, typename, membername, a, strlong);
		}

		public string MonoImpInfo(string assemblyname, string typename, string membername, ArrayList arglist, bool strlong) {
			try {
				Assembly assembly = Assembly.Load(assemblyname);
				if (assembly == null) {
					if (strlong) return "The assembly " + assemblyname + " is not available to MonoDoc.";
					else return "";
				}

				Type t = assembly.GetType(typename, false);
				if (t == null) {
					if (strlong) return typename + " has not been implemented.";
					else return "Not implemented.";
				}

				MemberInfo[] mis = t.GetMember(membername, BF.Static | BF.Instance | BF.Public | BF.NonPublic);

				if (mis.Length == 0) return "This member has not been implemented.";
				if (mis.Length == 1) return MonoImpInfo(mis[0], "member", strlong);

				// Scan for the appropriate member
				foreach (MemberInfo mi in mis) {
					System.Reflection.ParameterInfo[] pis;

					if (mi is MethodInfo || mi is ConstructorInfo)
						pis = ((MethodBase)mi).GetParameters();
					else if (mi is PropertyInfo)
						pis = ((PropertyInfo)mi).GetIndexParameters();
					else
						pis = null;
					
					if (pis != null) {
						bool good = true;
						if (pis.Length != arglist.Count) continue;
						for (int i = 0; i < pis.Length; i++) {
							if (pis[i].ParameterType.FullName != (string)arglist[i]) { good = false; break; }
						}
						if (!good) continue;
					}

					return MonoImpInfo(mi, "member", strlong);
				}

				return "This member has not been implemented.";

			} catch (Exception e) {
				return "An error occured while loading type information: " + e.Message;
			}
		}
		public string MonoImpInfo(System.Reflection.MemberInfo mi, string itemtype, bool strlong) {
			string s = "";

			object[] atts = mi.GetCustomAttributes(true);
			int todoctr = 0;
			foreach (object att in atts) if (att.GetType().Name == "MonoTODOAttribute") todoctr++;

			if (todoctr > 0) {
				if (strlong)
					s = "This " + itemtype + " is marked as being unfinished.<BR/>\n";
				else 
					s = "Unfinished.";
			}

			return s;
		}
		public string MonoImpInfo(string assemblyname, string typename, bool strlong) {
			try {
				if (assemblyname == "") return "";

				Assembly assembly = Assembly.Load(assemblyname);
				if (assembly == null) {
					if (strlong) return "The assembly " + assemblyname + " is not available to MonoDoc.";
					else return "";
				}

				Type t = assembly.GetType(typename, false);
				if (t == null) {
					if (strlong) return typename + " has not been implemented.";
					else return "Not implemented.";
				}

				string s = MonoImpInfo(t, "type", strlong);

				if (strlong) {
					MemberInfo[] mis = t.GetMembers(BF.Static | BF.Instance | BF.Public | BF.NonPublic);

					// Scan members for MonoTODO attributes
					int mctr = 0;
					foreach (MemberInfo mi in mis) {
						string mii = MonoImpInfo(mi, null, false);
						if (mii != "") mctr++; 
					}
					if (mctr > 0) {
						s += "This type has " + mctr + " members that are marked as unfinished.<BR/>";
					}
				}

				return s;

			} catch (Exception e) {
				return "An error occured while loading type information: " + e.Message;
			}			
		}
	}

	//
	// This takes one of the ecma urls, which look like this:
	// ecma:NUMERIC_ID#OPAQUE/REST
	//
	// NUMERIC_ID is the numeric ID assigned by the compressor
	// OPAQUE is opaque for node rendering (it typically contains T:System.Byte for example)
	// REST is the rest of the argument used to decode information
	//
	static string GetFile (string url, out string rest)
	{
		int pound = url.IndexOf ("#");
		int slash = url.IndexOf ("/");
		
		string fname = url.Substring (0, pound);
		rest = url.Substring (slash+1);

		return fname;
	}

	// This should have a little cache or something.
	static XmlDocument GetDocument (HelpSource hs, string fname)
	{
		Stream s = hs.GetHelpStream (fname);
		if (s == null){
			Console.Error.WriteLine ("Could not fetch document {0}", fname);
			return null;
		}
		
		XmlDocument doc = new XmlDocument ();

		doc.Load (s);
		
		return doc;
	}

	//
	// Populates the index.
	//
	public override void PopulateIndex (IndexMaker index_maker)
	{
		foreach (Node ns_node in Tree.Nodes){
			foreach (Node type_node in ns_node.Nodes){
				string typename = type_node.Caption.Substring (0, type_node.Caption.IndexOf (' '));
				string full = ns_node.Caption + "." + typename;

				if (type_node.Caption.EndsWith (" Class") ||
				    type_node.Caption.EndsWith (" Struct")){
					string url = "T:" + full;
					
					index_maker.Add (type_node.Caption, typename, url);
					index_maker.Add (full, full, url);
					
					foreach (Node c in type_node.Nodes){
						switch (c.Caption){
						case "Constructors":
							index_maker.Add ("  constructors", typename + "0", url + "/C");
							break;
						case "Fields":
							index_maker.Add ("  fields", typename + "1", url + "/F");
							break;
						case "Events":
							index_maker.Add ("  events", typename + "2", url + "/E");
							break;
						case "Properties":
							index_maker.Add ("  properties", typename + "3", url + "/P");
							break;
						case "Methods":
							index_maker.Add ("  methods", typename + "4", url + "/M");
							break;
						}
					}
				}
			}
		}
	}
}
